/**
 * \addtogroup Lcd
 * @{
 */

/*****************************************************************************/
/**
 * \file Lcd.c
 * HD44780 LCD display controller routines for 8051 and Keil C51.
 * \author Copyright (c) 2005, Murray R. Van Luyn. <vanluynm@iinet.net.au>
 * \version 0.1
 * \date 01-03-05
 */
/*****************************************************************************/

/*****************************************************************************
 *
 *     THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 *     OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *     WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *     ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 *     DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *     DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 *     GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *     INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *     WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *     NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *     SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 ******************************************************************************/

#include "Lcd.h"

/*****************************************************************************
 *
 *                         Private Function Prototypes
 *
 *****************************************************************************/
static void lcd_clock_E(void);
static void lcd_wr_reg(bit reg, unsigned char inst);
static unsigned char lcd_rd_reg(bit reg);
static void lcd_backspace(void);
static void lcd_clear(void);
static void lcd_scroll_up(void);

/*****************************************************************************
 *
 *                        Private Function Implementation
 *
 *****************************************************************************/

/*****************************************************************************
 *
 *                                lcd_clock_E()
 *
 *****************************************************************************/
static void lcd_clock_E(void)
{
    E = 1;
    delay_ms(0);               // delay > 100us
    delay_ms(0);
    delay_ms(0);
    E = 0;
}

/*****************************************************************************
 *
 *                                lcd_wr_reg()
 *
 *****************************************************************************/
static void lcd_wr_reg(bit reg, unsigned char inst)
{
    RS = reg;                  // Select instruction or data register
    RW = 0;                    // Select write operation
    LCD_PORT &= 0xF0;          // Clear port data
    LCD_PORT |= inst >> 4;     // Send D7..D4 to lcd
    lcd_clock_E();
    LCD_PORT &= 0xF0;          // Clear port data
    LCD_PORT |= inst & 0x0F;   // Send D3..D0 to lcd
    lcd_clock_E();
    LCD_PORT |= 0x0F;          // Set low nibble for input
                               // Prevent damage
}

/*****************************************************************************
 *
 *                                lcd_rd_reg()
 *
 *****************************************************************************/
static unsigned char lcd_rd_reg(bit reg)
{
    unsigned char value;

    LCD_PORT |= 0x0F;          // Set low nibble for input
    RS = reg;                  // Select instruction or data register
    RW = 1;                    // Select read operation
    delay_ms(0);               // Wait 40 ns minimum 
    E = 1;                     // Enable output
    delay_ms(0);               // Wait 160 ns minimum 
    value = LCD_PORT << 4;     // Retrieve D7..D4
    E = 0;                     // Disable output
    delay_ms(0);               // Wait 230 ns minimum 
    E = 1;                     // Enable output
    delay_ms(0);               // Wait 160 ns minimum 
    value |= LCD_PORT & 0x0F;  // Retrieve D3..D0
    E = 0;                     // Disable output
    RW = 0;                    // Prevent damage

    return(value);       
}

/*****************************************************************************
 *
 *                                lcd_backspace()
 *
 *****************************************************************************/
static void lcd_backspace(void)
{
    unsigned char addr;

    addr = LCD_GET_ADDR();

    if ((addr != 0x00) && (addr != 0x40))
    {
        LCD_SET_ADDR(addr - 1);
    }
}

/*****************************************************************************
 *
 *                                lcd_clear() 
 *
 *****************************************************************************/
static void lcd_clear(void)
{
    lcd_wr_reg(INST_REG, CLEAR_DISPLAY);

    delay_ms(2);               // Wait more than 1.64ms
}

/*****************************************************************************
 *
 *                               lcd_scroll_up() 
 *
 *****************************************************************************/
static void lcd_scroll_up(void)
{
    unsigned char i;
    unsigned char c;

    for(i = LINE1_ADDR; i < 0x28; i++)
    {
        LCD_SET_ADDR(i + LINE2_ADDR);
        c = lcd_rd_reg(DATA_REG);
        LCD_SET_ADDR(i + LINE2_ADDR);
        lcd_wr_reg(DATA_REG, ' ');
        LCD_SET_ADDR(i);
        lcd_wr_reg(DATA_REG, c);
    }
    LCD_SET_ADDR(LINE2_ADDR);
}


/*****************************************************************************
 *
 *                         Public Function Implementation
 *
 *****************************************************************************/

/*****************************************************************************
 *
 *                                 init_lcd()
 *
 *****************************************************************************/
void lcd_init(void)
{
    delay_ms(15);                // Wait more than 15ms
    LCD_PORT = 0x03;             // Startup Sequence
    lcd_clock_E();
    delay_ms(5);                 // Wait more than 4.1ms
    lcd_clock_E();  
    delay_ms(1);                 // Wait more than 0.1ms
    lcd_clock_E();    
    LCD_PORT = 0x02;
    lcd_clock_E();
    delay_ms(1);                 // Wait more than 0.1ms

    lcd_wr_reg(INST_REG,0x28);   // Function Set
                                 // DL=0 4bit, N=1 2Line, F=0 5x7

    lcd_wr_reg(INST_REG,0x0c);   // Display on/off control
                                 // D=1 Disp on, C=0 Curs off, B=0 Blink off

    lcd_wr_reg(INST_REG,0x06);   // Entry Mode Set
                                 // I/D=1 Increment, S=0 No shift

    lcd_wr_reg(INST_REG,0x01);   // Clear Display

    delay_ms(2);                 // Wait more than 1.64ms

}

/*****************************************************************************
 *
 *                                 putchar()
 *
 *****************************************************************************/
char putchar(char c)
{
    switch(c)
    {
        case '\n':           // Newline
        {
            if (LCD_GET_ADDR() >= LINE2_ADDR)    // already in line 2 ?
            {
                lcd_scroll_up();
            }
            LCD_SET_ADDR(LINE2_ADDR);
            break;
        }
        case '\a':           // Bell (alert)
        {
            lcd_beep(50);
            break;
        }
        case '\b':           // Backspace
        {
            lcd_backspace();
            break;
        }
        case '\f':           // Form feed
        {
            lcd_clear();
            break;
        }
        default:
        {
            lcd_wr_reg(DATA_REG, c);
        }
    }
    return (c);
}

/*****************************************************************************
 *
 *                                    lcd_beep()
 *
 *****************************************************************************/
void lcd_beep(unsigned char count)
{
    for (count; count > 0; count--)
    {
        SPKR = 1;
        delay_ms(1);
        SPKR = 0;
        delay_ms(1);
    }
}

/** @} */
